import java.util.*;
import java.util.stream.Collectors;

/**
 * Created by hanngu on 18.02.15.
 */
public class ACO {
    private Random rand = new Random();
    protected Neo4jHandler neo4jHandler;
    protected int numberOfIterations;
    protected int iterationNumber;
    protected Statistics statistics;
    protected Ant bestAnt;
    protected ArrayList<MandlNodeCouple> nodeCouples;
    protected ArrayList<MandlNode> nodes;
    protected ArrayList<MandlEdge> mandlEdges;
    protected double pheromoneVisited, pheromoneBest, totalDemand;


    public ACO(Neo4jHandler neo4jHandler, Statistics statistics, int numberOfIterations, double pheromoneVisited, double pheromoneBest, double totalDemand) {
        this.statistics = statistics;
        this.neo4jHandler = neo4jHandler;
        this.numberOfIterations = numberOfIterations;
        this.pheromoneVisited = pheromoneVisited;
        this.pheromoneBest = pheromoneBest;
        this.totalDemand = totalDemand;
        this.iterationNumber = 0;
        nodeCouples = new ArrayList<>();
        nodes = new ArrayList<>();
        mandlEdges = new ArrayList<>();


    }

    public ACO(double totalDemand){
        this.totalDemand = totalDemand;
        nodeCouples = new ArrayList<>();
        nodes = new ArrayList<>();
        mandlEdges = new ArrayList<>();
    }

    public ArrayList<Ant> generateAnts(int colonySize, ArrayList<MandlNode> mandlNodes, int maxNodesInRoute, int minNodesInRoute) {
        iterationNumber ++;
        this.nodes = mandlNodes;
        ArrayList<Ant> ants = new ArrayList<>();
        for (int i = 0; i < colonySize; i++) {
            MandlNode firstNode = chooseFirstNode();
            boolean crazyAnt = false;
            Ant ant = new Ant(i, firstNode, maxNodesInRoute, crazyAnt, minNodesInRoute);
            ants.add(ant);
        }

        return ants;
    }

    public Ant getBestAnt() {
        return this.bestAnt;
    }

    public void setNodeCouples(ArrayList<MandlNodeCouple> nodeCouples) {
        this.nodeCouples = nodeCouples;
    }

    public void createRoutes(int numberOfRoutes, ArrayList<Ant> ants, ArrayList<MandlEdge> mandlEdges) {
        this.mandlEdges = mandlEdges;
        //For hver maur i ants, så skal det genereres et gitt antall ruter.
        for (Ant a : ants) {
            if(a.isFollower()){
                continue;
            }
            //Generates nodeValues based on the number of edges. Nodes with only one edge should be unfavored when when creating routes, because they make the ant stuck.
            for (int i = 0; i < numberOfRoutes; i++) {
                boolean keepCreatingRoutes = true;
                while (keepCreatingRoutes) {
                    keepCreatingRoutes = chooseNextNodes(a);
                }
            }
        }

    }

    //Evalueringen tar 10 av 18 sekunder.
    public void evaluation(ArrayList<Ant> ants) {
        ArrayList<Ant> bestAnts = new ArrayList<>();
        ArrayList<Integer> nodeNumbers = new ArrayList<>();
        for (MandlNode n : nodes) nodeNumbers.add(n.getNumber());
        for (Ant a : ants) {
            ArrayList<Integer> copyOfNodeNumbers = new ArrayList<>();
            for (Integer i : nodeNumbers)
                copyOfNodeNumbers.add(i);

            for (ArrayList<Integer> route : a.getRouteSet()) {
                if (!copyOfNodeNumbers.isEmpty()) {
                    for (Integer n : nodeNumbers) {
                        if (route.contains(n) && copyOfNodeNumbers.contains(n)) {
                            copyOfNodeNumbers.remove(n);
                        }
                    }
                }
            }
            if (copyOfNodeNumbers.isEmpty()) {
                bestAnts.add(a);
            }
        }
        //Generates couple of mandlNodes, that should be used in evaluation. The number of node1 is at [0] in the array, and the number of node2 is at [1] in the array
        //ArrayList<MandlNodeCouple> totalNodeCouples = createNodeCouples(mandlNodes, neo4jHandler);
        ArrayList<Ant> disconnectedAnts = new ArrayList<>();
        double totalFitOfAllAnts = 0.0;
        for (Ant a : bestAnts) {
            //Calculates F1
            //Calculates the dif between shortest path i the network and the shortest path possible given the route set of ant a for every nodeCouple in nodeCouples. This number should be as small as possible
            int dif = 0;
            dif += neo4jHandler.findShortestPathForCurrentAnt(a, nodeCouples, iterationNumber);
            if (dif == -1) {
                disconnectedAnts.add(a);
                continue;
            }
            a.setF1(dif);
            //Calculates F2
            //Calculates a score based on number of direct travelers, travelers with only one transfer and travelers with only two transfers. The returned value is the number of MandlNodeCouples that fulfills these criterias
            double satisfiedCouples = calculateF2(a, nodeCouples);

            //Calculates F3
            //Adds the number of unsatisfied MandleNodeCouples (totalNodeCouples-satisfiedCouples) to the ant. A good ant has as few as possible of these.
            double unsatisfied = nodeCouples.size() - satisfiedCouples;
            a.setF3(unsatisfied);
            totalFitOfAllAnts += a.getTotaltFit();

        }
        //The java 8 way
        disconnectedAnts.forEach(bestAnts::remove);
        if(bestAnts.size() == 0){
            System.out.println("No ants satisfied the initial demands");
            System.exit(0);
        }

        Collections.sort(bestAnts, new TotfitComparator());
        int numberOfAntsToBeChosen = bestAnts.size() / 2; //If the the size is an odd number, it will not include the middel one
        ArrayList<Ant> finalBestAnts = new ArrayList<>();
        if (bestAnt == null || bestAnt.getTotaltFit() > bestAnts.get(0).getTotaltFit()) {
            bestAnt = bestAnts.get(0);
        }
        for (int i = 0; i < numberOfAntsToBeChosen; i++) {
            finalBestAnts.add(bestAnts.get(i));
        }

        Collections.sort(finalBestAnts, new DemandComparator());
        numberOfAntsToBeChosen = finalBestAnts.size() / 2;
        ArrayList<Ant> bestAntsBasedOnDemand = new ArrayList<>();
        for (int i = 0; i < numberOfAntsToBeChosen; i++) {
            bestAntsBasedOnDemand.add(finalBestAnts.get(i));
        }

        ArrayList<MandlEdge> edgesThatAreGettingMorePheromone = findFinalBestAntsEdges(bestAntsBasedOnDemand);
        giveBestRouteSetsMorePheromone(edgesThatAreGettingMorePheromone);

    }

    protected void calculateDemandSatisfiedByDirectTravel(Ant a) {
        double demandSatisfiedByDirectTravel = 0;
        for(MandlNodeCouple nodeCouple: a.getDirectCouples()){
            demandSatisfiedByDirectTravel += nodeCouple.getDemand();
        }
        a.setDemandSatisfiedByDirectTravel(demandSatisfiedByDirectTravel);
    }

    protected void calculateDemandSatisfiedByOneTransfer(Ant a){
        double demandSatisfiedByOneTransfer = 0;
        for(MandlNodeCouple nodeCouple: a.getOneTransferCouples()){
            demandSatisfiedByOneTransfer += nodeCouple.getDemand();
        }
        a.setDemandSatisfiedByOneTransfer(demandSatisfiedByOneTransfer);
    }

    protected void calculateDemandSatisfiedByTwoTransfers(Ant a){
        double demandSatisfiedByTwoTransfers = 0;
        for(MandlNodeCouple nodeCouple: a.getTwoTransferCouples()){
            demandSatisfiedByTwoTransfers += nodeCouple.getDemand();
        }
        a.setDemandSatisfiedByTwoTransfers(demandSatisfiedByTwoTransfers);
    }

    protected ArrayList<MandlEdge> findFinalBestAntsEdges(ArrayList<Ant> finalBestAnts) {
        ArrayList<MandlEdge> edgesToReturn = new ArrayList<>();
        for (Ant a : finalBestAnts) {
            edgesToReturn.addAll(a.getEdgesInRouteSet().stream().collect(Collectors.toList()));
        }
        return edgesToReturn;
    }

   public double calculateF2(Ant ant, ArrayList<MandlNodeCouple> nodeCouples) {
       ArrayList<MandlNodeCouple> directCouples = new ArrayList<>();
       ArrayList<MandlNodeCouple> oneTransferCouples = new ArrayList<>();
       ArrayList<MandlNodeCouple> twoTransferCouples = new ArrayList<>();
        int numberOfNodeCouples = nodeCouples.size();
        //Works according to preelimnary testing
       directCouples = ant.getDirectCouples();
       if(directCouples.size() == nodeCouples.size()){
           System.out.println("bro");
           System.out.println(ant.isCrazyAnt());
           System.out.println();
       }

        //Works according to preelimnary testing
       oneTransferCouples = ant.getOneTransferCouples();


        if((directCouples.size() + oneTransferCouples.size()) != numberOfNodeCouples){
            twoTransferCouples = ant.getTwoTransferCouples();
        }



        calculateDemandSatisfiedByDirectTravel(ant);
        calculateDemandSatisfiedByOneTransfer(ant);
        calculateDemandSatisfiedByTwoTransfers(ant);
        double demandNotSatisfied = totalDemand - ant.getDemandSatisfiedByDirectTravel() - ant.getDemandSatisfiedByOneTransfer() - ant.getDemandSatisfiedByTwoTransfers();
        ant.setDemandNotSatisfied(demandNotSatisfied);

        //ant.setTotalTravelTime(ant.getTotalTravelTime() + (ant.getDemandSatisfiedByOneTransfer() * 5));
        //ant.setTotalTravelTime(ant.getTotalTravelTime() + (ant.getDemandSatisfiedByTwoTransfers() * 10));

        ant.setD0((ant.getDemandSatisfiedByDirectTravel() / totalDemand) * 100);
        ant.setD1((ant.getDemandSatisfiedByOneTransfer() / totalDemand)  * 100);
        ant.setD2((ant.getDemandSatisfiedByTwoTransfers() / totalDemand) * 100);
        ant.setdUnsat(100 - ant.getD0() - ant.getD1() - ant.getD2());

        //We must appriciate ants with many direct routes. This may be done by weighting direct routes with w1=-3, routes with two transfers with w2=-2 and just let the number of travelers with two transfers with w3=-1.
        double f2 = ((-3) * ant.getD0()) + ((-2) * ant.getD1()) + ((-1) * ant.getD2());
        //System.out.println("f2: " + f2);
        ant.setF2(f2);

        return directCouples.size() + oneTransferCouples.size() + twoTransferCouples.size();
    }

    //Removes a bit of pheromone as time goes by
    public void evaporation(double evaporationDelta) {
        double pheromoneToRemain = 1.0-evaporationDelta;
        for(MandlEdge edge : mandlEdges){
            if(edge.getTravelTime() != -1){
                edge.decreasePheromone(pheromoneToRemain);
            }
        }

    }


    private boolean chooseNextNodes(Ant a) {
        //Special case: If the node to be chosen is the first i the route
        if (a.getCurrentRoute().size() == 0) {
            MandlNode chosenNode = chooseFirstNode();
            a.setNextNode(chosenNode, neo4jHandler, iterationNumber);
            return true;
        }
        ArrayList<MandlNode> possibleMandlNodes = new ArrayList<>();
        ArrayList<MandlEdge> possibleMandlEdges = a.getCurrentMandlNode().getConnectingEdges();
        for (MandlEdge e : possibleMandlEdges) {
            //If the the ants current node is equal to the edge's Node1, then Node2 should be added as a possibleNode.
            if (e.getMandlNode1() == a.getCurrentMandlNode()) {
                possibleMandlNodes.add(e.getMandlNode2());
            } else {
                possibleMandlNodes.add(e.getMandlNode1());
            }
        }
        possibleMandlEdges = removeImpossibleEdges(possibleMandlEdges, a); //removes mandlEdges that contains mandlNodes that are already in ant a's route set
        if (possibleMandlEdges.isEmpty()) { //if the ant is stuck, it is stuck
            a.finishedRoute(neo4jHandler, iterationNumber);
            return false;
        }

        MandlEdge chosenMandlEdge = selectEdge(possibleMandlEdges, a);
        chosenMandlEdge.increasePheromone(pheromoneVisited);

        boolean routeSetsNotFull = selectNextNode(chosenMandlEdge, a);
        if (routeSetsNotFull) {
            a.addEdge(chosenMandlEdge);
        }
        return routeSetsNotFull;

    }

    private ArrayList<MandlEdge> removeImpossibleEdges(ArrayList<MandlEdge> possibleMandlEdges, Ant a) {
        ArrayList<MandlEdge> actualPossibleMandlEdges = new ArrayList<>();
        for (MandlEdge e : possibleMandlEdges) {
            if (e.getMandlNode1() == a.getCurrentMandlNode()) {
                if (!a.getCurrentRoute().contains(e.getMandlNode2().getNumber())) {
                    actualPossibleMandlEdges.add(e);
                }

            } else {
                if (!a.getCurrentRoute().contains(e.getMandlNode1().getNumber())) {
                    actualPossibleMandlEdges.add(e);
                }
            }
        }
        return actualPossibleMandlEdges;
    }

    private boolean selectNextNode(MandlEdge chosenMandlEdge, Ant a) {
        MandlNode nextMandlNode;
        if (chosenMandlEdge.getMandlNode1() == a.getCurrentMandlNode()) {
            nextMandlNode = chosenMandlEdge.getMandlNode2();
        } else {
            nextMandlNode = chosenMandlEdge.getMandlNode1();
        }
        return a.setNextNode(nextMandlNode, neo4jHandler, iterationNumber);
    }

    //TODO: Make prettier
    //Ugly, but it works
    protected MandlEdge selectEdge(ArrayList<MandlEdge> possibleMandlEdges, Ant a) {
        double totalPheromone = 0.0;
        double totalDemand = 0.0;
        double totalUnexploredNodes = 0.0;


        for (MandlEdge e : possibleMandlEdges) {
            totalPheromone += e.getPheromone();
            totalDemand += e.getDemand();
            //totalUnexploredNodes will be the number of nodes in possible edges that have not been visited by ant a in any of its route sets
            if (e.getMandlNode1().getNumber() == a.getCurrentMandlNode().getNumber()) {
                if (checkIfNodeIsInAntsRouteSet(a, e.getMandlNode2())) {
                    totalUnexploredNodes++;
                }
            } else {
                if (checkIfNodeIsInAntsRouteSet(a, e.getMandlNode1())) {
                    totalUnexploredNodes++;
                }
            }
        }
        if (totalPheromone == 0) {
            totalPheromone = -1;
        }
        if (totalDemand == 0) {
            totalDemand = -1;
        }
        if (totalUnexploredNodes == 0) {
            totalUnexploredNodes = -1;
        }

        double total = 0;

        for (MandlEdge e : possibleMandlEdges) {
            double unexploredNode = 0;

            if (e.getMandlNode1().getNumber() == a.getCurrentMandlNode().getNumber()) {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode2())) {
                    unexploredNode = 1;
                }
            } else {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode1())) {
                    unexploredNode = 1;
                }
            }
            double edgeValue = Math.abs((e.getDemand() / totalDemand) + (e.getPheromone() / totalPheromone) + (unexploredNode / totalUnexploredNodes));
            total += edgeValue;
            e.setEdgeValue(edgeValue);


        }
        double staticTotal = total;
        if (total == 0) {
            double tempValue = 1.0;
            boolean firstIter = true;
            for (MandlEdge e : possibleMandlEdges) {
                if (!firstIter) {
                    tempValue -= 1 / possibleMandlEdges.size();
                }
                if(e.getProbability() == 0.0){
                    e.setProbability(tempValue);
                }
                total += 1 / possibleMandlEdges.size();
                firstIter = false;
            }
        } else {
            boolean first = true;
            for (MandlEdge e : possibleMandlEdges) {
                double range = total;
                if (!first) {
                    range = total - e.getEdgeValue();
                    total -= e.getEdgeValue();
                }
                double probability = range;
                e.setProbability(probability);
                first = false;
            }

        }
        Random r = new Random();
        double randis = 0 + (staticTotal - 0) * r.nextDouble();
        MandlEdge chosenMandlEdge = null;
        Collections.reverse(possibleMandlEdges);
        for (MandlEdge e : possibleMandlEdges) {
            if (randis < e.getProbability()) {
                chosenMandlEdge = e;
                break;
            }
        }
        for (MandlEdge e : possibleMandlEdges) {
            e.setProbability(0.0);
            e.setEdgeValue(0.0);
        }
        ArrayList<MandlEdge> edgesWithUnexploredNodes = new ArrayList<>();
        for (MandlEdge e : possibleMandlEdges) {
            if (e.getMandlNode1().getNumber() == a.getCurrentMandlNode().getNumber()) {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode2())) {
                    edgesWithUnexploredNodes.add(e);
                }
            } else {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode1())) {
                    edgesWithUnexploredNodes.add(e);
                }
            }
        }
        if (!edgesWithUnexploredNodes.isEmpty()) {
            if (edgesWithUnexploredNodes.size() != 1) {
                chosenMandlEdge = edgesWithUnexploredNodes.get(0);
            }
            if (edgesWithUnexploredNodes.size() == 1) {
                chosenMandlEdge = edgesWithUnexploredNodes.get(0);

            }
        }
        return chosenMandlEdge;
    }



    protected boolean checkIfNodeIsInAntsRouteSet(Ant a, MandlNode mandlNode2) {

        for (ArrayList<Integer> route : a.getRouteSet()) {
            for (int i : route) {
                if (i == mandlNode2.getNumber()) {
                    return true;
                }
            }
        }
        return false;

    }

    protected MandlNode chooseFirstNode() {
        boolean validNode = false;
        MandlNode choosenNode = null;
        while (!validNode) {
            int randomNumber = rand.nextInt((nodes.size()));
            MandlNode nodeToBeTested = nodes.get(randomNumber);
            if (checkIfNodeIsASuitableFirstNode(nodeToBeTested)) {
                choosenNode = nodeToBeTested;
                validNode = true;
            }
        }
        return choosenNode;
    }

    private boolean checkIfNodeIsASuitableFirstNode(MandlNode nodeToBeTested) {
        ArrayList<MandlNode> connectingNodes = new ArrayList<>();
        for (MandlEdge edge : nodeToBeTested.getConnectingEdges()) {
            if (edge.getMandlNode1().getNumber() != nodeToBeTested.getNumber()) {
                connectingNodes.add(edge.getMandlNode1());
            } else {
                connectingNodes.add(edge.getMandlNode2());
            }
        }
        for(MandlNode node : connectingNodes){
            if(node.getConnectingEdges().size() == 1){
                return false;
            }
        }
        return true;
    }

    public void giveBestRouteSetsMorePheromone(ArrayList<MandlEdge> mandlEdges) {

        for (MandlEdge edge : mandlEdges) {
            edge.increasePheromone(pheromoneBest);
        }

    }
}
