import org.neo4j.graphalgo.*;
import org.neo4j.graphdb.*;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.factory.GraphDatabaseFactory;
import org.neo4j.graphdb.index.Index;
import org.neo4j.graphdb.index.IndexManager;
import org.neo4j.io.fs.FileUtils;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Created by hanngu on 25.02.15.
 */
public class Neo4jHandler {

    private GraphDatabaseService graphDb;
    private static String dataDirectory;
    private IndexManager index;
    private Index<Node> nodes;
    private Index<Relationship> relationships;
    private ArrayList<MandlEdge> edgesWithValidTravelTime;
    private ArrayList<MandlEdge> mandlEdges;
    private long timer = 0;
    private RelationshipType demandRelationship = new RelationshipType() {
        @Override
        public String name() {
            return "hasDemand";
        }
    };

    public Neo4jHandler (String filename) {
        this.dataDirectory = filename;
        this.graphDb = new GraphDatabaseFactory().newEmbeddedDatabase(filename);
        registerShutdownHook(this.graphDb);
    }

    public double getAverageDemand() {
        return averageDemand;
    }

    private double averageDemand = 0.0;

    public File getStorageDirectory() throws NoSuchFieldException {
        return new File(String.valueOf(graphDb.getClass().getDeclaredField("storeDir")));
    }

    public void cleanUp() {
        try {
            File data = new File(dataDirectory);
            assert data.isDirectory();
            FileUtils.deleteRecursively(data);
        } catch (IOException e) {
            System.out.println("INSIDE CLEANUPS IO EXCEPTION");
            e.printStackTrace();
        }

    }
    public GraphDatabaseService createNeo4jDatabase(GenerateNetwork gn){
        ArrayList<MandlNode> mandlNodes = gn.getMandlNodes();
        mandlEdges = gn.getMandlEdges();

        edgesWithValidTravelTime = new ArrayList<MandlEdge>();
        for(MandlEdge edge : mandlEdges){
            if(edge.getTravelTime() > 0){
                edgesWithValidTravelTime.add(edge);
            }
        }
        try ( Transaction tx = graphDb.beginTx() )
        {
            index = graphDb.index();
            nodes = index.forNodes("nodes");
            for(MandlNode mandlNode : mandlNodes){
                Node node = graphDb.createNode();
                node.addLabel(DynamicLabel.label("Node"));
                node.setProperty("xCord", mandlNode.getCordX());
                node.setProperty("yCord", mandlNode.getCordY());
                node.setProperty("number", mandlNode.getNumber());
                nodes.add(node, "xCord", node.getProperty("xCord"));
                nodes.add(node, "yCord", node.getProperty("yCord"));
                nodes.add(node, "number", node.getProperty("number"));
                //System.out.println(node.getProperty("number"));

                //System.out.println(node.getProperty("number") + "," + node.getProperty("xCord") + "," + node.getProperty("yCord"));
            }
            //Creates the actual possible edges
            relationships = index.forRelationships("edges");
            for(MandlEdge edge : edgesWithValidTravelTime){
                MandlNode mandlNode1 = edge.getMandlNode1();
                MandlNode mandlNode2 = edge.getMandlNode2();
                Node node1 = nodes.get("number", mandlNode1.getNumber()).getSingle();
                Node node2 = nodes.get("number", mandlNode2.getNumber()).getSingle();
                Relationship relationship = node1.createRelationshipTo(node2, RelTypes.HAS_CONNECTION);
                relationship.setProperty("travelTime", edge.getTravelTime());
                //System.out.println(node1.getProperty("number") + "-[" + relationship.getProperty("travelTime") + "]->" + node2.getProperty("number"));
            }
            //Every node couple has a demand. This must be represented.
            for(MandlEdge edge : mandlEdges) {
                MandlNode mandlNode1 = edge.getMandlNode1();
                MandlNode mandlNode2 = edge.getMandlNode2();
                Node node1 = nodes.get("number", mandlNode1.getNumber()).getSingle();
                Node node2 = nodes.get("number", mandlNode2.getNumber()).getSingle();
                Relationship relationship = node1.createRelationshipTo(node2, demandRelationship);
                relationship.setProperty("demand", edge.getDemand());
                //System.out.println(node1.getProperty("number") + "-[" + relationship.getProperty("demand") + "]->" + node2.getProperty("number"));

            }

            generateAverageDemand(mandlEdges);
            tx.success();
        }

        return graphDb;

    }

    private void generateAverageDemand(ArrayList<MandlEdge> mandlEdges) {
        int totalDemand = 0;
        int totalNodeCouples = 0;
        for(MandlEdge edge : mandlEdges){
            totalDemand += edge.getDemand();
            totalNodeCouples ++;
        }
        averageDemand = totalDemand/totalNodeCouples;
        //System.out.println("Average demand: " + averageDemand);
    }

    public double findShortesPathBetweenTwoNodes(MandlNode mandlNode1, MandlNode mandlNode2){

        try ( Transaction tx = graphDb.beginTx() )
        {
            Node node1 = nodes.get("number", mandlNode1.getNumber()).getSingle();
            Node node2 = nodes.get("number", mandlNode2.getNumber()).getSingle();

            RelationshipType type1 = new RelationshipType() {
                @Override
                public String name() {
                    return "HAS_CONNECTION";
                }
            };
            PathFinder<WeightedPath> dijkstra = GraphAlgoFactory.dijkstra(
                    PathExpanders.forType(type1),
                    CommonEvaluators.doubleCostEvaluator("travelTime"));
            WeightedPath path = dijkstra.findSinglePath(node1, node2);

            tx.success();
            return path.weight();
        }

    }

    public void addRouteToGraph(ArrayList<Integer> currentRoute, int routeNumber, int antNumber, int iterationNumber) {

        try ( Transaction tx = graphDb.beginTx() )
        {
            for(int i = 0; i < currentRoute.size()-1; i ++){
                Node node1 = nodes.get("number", currentRoute.get(i)).getSingle();
                Node node2 = nodes.get("number", currentRoute.get(i+1)).getSingle();
                RelationshipType type = new RelationshipType() {
                    @Override
                    public String name() {
                        return antNumber+"-"+routeNumber+"-"+iterationNumber;
                    }
                };
                Relationship relationship = node1.createRelationshipTo(node2, type);
                relationship.setProperty("type", type.name());
                for(MandlEdge edge : edgesWithValidTravelTime){
                    if((edge.getMandlNode1().getNumber() == currentRoute.get(i) &&  edge.getMandlNode2().getNumber() == currentRoute.get(i+1)) ||
                            (edge.getMandlNode2().getNumber() == currentRoute.get(i) &&  edge.getMandlNode1().getNumber() == currentRoute.get(i+1))){

                        relationship.setProperty("travelTime", edge.getTravelTime());

                        //System.out.println("Node1:" + currentRoute.get(i) + ", Node2: " + currentRoute.get(i+1) + ", TravelTime: " + edge.getTravelTime() + ", Type: " + type.name());
                    }

                }

            }
            tx.success();
        }

    }

    public double findShortestPathForCurrentAnt(Ant a, ArrayList<MandlNodeCouple> nodeCouples, int iterationNumber) {
        ArrayList<MandlNodeCouple> directCouples = new ArrayList<>();
        ArrayList<MandlNodeCouple> oneTransferCouples = new ArrayList<>();
        ArrayList<MandlNodeCouple> twoTransferCouples = new ArrayList<>();
        ArrayList<MandlNodeCouple> unsatisfiedCouples = new ArrayList<>();
        int routeSetSize = a.getRouteSet().size();
        int dif = 0;
        double totalTravelTimeForTheEntireNetwork = 0.0;
        ArrayList<Object> relationShipsAndDirections = new ArrayList<Object>();

        //Heuristic function? used by astar
        EstimateEvaluator<Double> estimateEvaluator = new EstimateEvaluator<Double>()
        {
            @Override
            public Double getCost( final Node node, final Node goal )
            {
                double dx = (Double) node.getProperty( "xCord" ) - (Double) goal.getProperty( "xCord" );
                double dy = (Double) node.getProperty( "yCord" ) - (Double) goal.getProperty( "yCord" );
                double result = Math.sqrt( Math.pow( dx, 2 ) + Math.pow( dy, 2 ) );
                return result;
            }
        };

        try ( Transaction tx = graphDb.beginTx() ) {
            Direction direction = Direction.BOTH;
            for (int i = 0; i < routeSetSize; i++) {
                final int finalI = i + 1;
                RelationshipType type = new RelationshipType() {
                    @Override
                    public String name() {
                        return a.getAntNumber() + "-" + finalI +"-" + iterationNumber;
                    }
                };

                relationShipsAndDirections.add(type);
                relationShipsAndDirections.add(direction);
            }
            int count = 0;
            for(MandlNodeCouple nodeCouple : nodeCouples){
                Node node1 = nodes.get("number", nodeCouple.getNode1().getNumber()).getSingle();
                Node node2 = nodes.get("number", nodeCouple.getNode2().getNumber()).getSingle();
                PathFinder<WeightedPath> aStar = GraphAlgoFactory.aStar(PathExpanders.forTypesAndDirections((RelationshipType) relationShipsAndDirections.get(0), (Direction) relationShipsAndDirections.get(1),
                        (RelationshipType) relationShipsAndDirections.get(2), (Direction) relationShipsAndDirections.get(3),
                        relationShipsAndDirections.size() > 3 ? relationShipsAndDirections.subList(4, relationShipsAndDirections.size()).toArray()
                                : new Object[0]), CommonEvaluators.doubleCostEvaluator("travelTime"), estimateEvaluator);
                PathFinder<WeightedPath> dijkstra = GraphAlgoFactory.dijkstra(
                        PathExpanders.forTypesAndDirections((RelationshipType) relationShipsAndDirections.get(0), (Direction) relationShipsAndDirections.get(1),
                                (RelationshipType) relationShipsAndDirections.get(2), (Direction) relationShipsAndDirections.get(3),
                                relationShipsAndDirections.size() > 3 ? relationShipsAndDirections.subList(4, relationShipsAndDirections.size()).toArray() : new Object[0]),
                        CommonEvaluators.doubleCostEvaluator("travelTime"));

                //WeightedPath path = dijkstra.findSinglePath(node1, node2);
                Iterable<WeightedPath> paths = dijkstra.findAllPaths(node1, node2);

                int counte = 0;
                ArrayList<PathInRouteSet> listOfPaths = new ArrayList<>();
                for(WeightedPath path : paths){
                    counte ++;
                    ArrayList<String> listOfRelationships = new ArrayList<>();
                    for(Relationship r : path.relationships()){
                        if(listOfRelationships.size() == 0 || !listOfRelationships.contains(r.getType().name())){
                            listOfRelationships.add(r.getType().name());
                        }
                    }
                    int numberOfRelationships = listOfRelationships.size();
                    double travelTime = path.weight() * nodeCouple.getDemand();
                    if(numberOfRelationships > 1){
                        travelTime += ((numberOfRelationships-1)*5*nodeCouple.getDemand());
                    }
                    PathInRouteSet pathInRouteSet = new PathInRouteSet(path, travelTime, numberOfRelationships);
                    listOfPaths.add(pathInRouteSet);
                }
                Collections.sort(listOfPaths, new TravelTimeComparator());
                if(listOfPaths.size() == 0){
                    //System.out.println("Node1: " + nodeCouple.getNode1().getNumber() + ", Node2: " + nodeCouple.getNode2().getNumber());
                    a.setTotalTravelTime(-1);
                    return -1;
                }
                PathInRouteSet chosenPath = listOfPaths.get(0);

                if(chosenPath.getNumberOfRelationships() == 0){
                    System.err.println("List of relationships = 0, should never happend");
                } else if(chosenPath.getNumberOfRelationships() == 1){
                    directCouples.add(nodeCouple);
                } else if(chosenPath.getNumberOfRelationships() == 2){
                    oneTransferCouples.add(nodeCouple);
                } else if(chosenPath.getNumberOfRelationships() == 3){
                    twoTransferCouples.add(nodeCouple);
                } else {
                    unsatisfiedCouples.add(nodeCouple);
                }

                double totalTravelTime = chosenPath.getPath().weight()*nodeCouple.getDemand();
                //System.out.println((totalTravelTime/nodeCouple.getMinTotalTravelTime()));
                /*System.out.println("Travel time chosen path: " + chosenPath.getTravelTime());
                System.out.println("Travel time best: " + nodeCouple.getMinTotalTravelTime());
                System.out.println();*/
                dif+= (chosenPath.getTravelTime() - (nodeCouple.getMinTotalTravelTime()*nodeCouple.getDemand())/nodeCouples.size());
                totalTravelTimeForTheEntireNetwork += totalTravelTime;
                tx.success();
            }


            a.setTotalTravelTime(totalTravelTimeForTheEntireNetwork);
            a.setDirectCouples(directCouples);
            a.setOneTransferCouples(oneTransferCouples);
            a.setTwoTransferCouples(twoTransferCouples);
            a.setUnsatisfiedCouples(unsatisfiedCouples);
            return dif;

        }



    }

    public long getTimer(){
        return timer;
    }

    public double calculateAverageDemandBetweenNodesForRouteSet(Ant a) {
        double totalDemand = 0;
        double numberOfNodeCouples = 0;
        try ( Transaction tx = graphDb.beginTx() ) {
            for (ArrayList<Integer> route : a.getRouteSet()) {
                for (int i = 1; i < route.size() - 1; i++) {
                    Node node1 = nodes.get("number", route.get(i)).getSingle();
                    Node node2 = nodes.get("number", route.get(i + 1)).getSingle();

                    for (Relationship r : node1.getRelationships(demandRelationship)){
                        if(r.getOtherNode(node1).equals(node2)){
                            String temp = ""+ r.getProperty("demand");
                            totalDemand += Double.parseDouble(temp);
                            numberOfNodeCouples ++;
                        }
                    }
                }
            }
            tx.success();
            //System.out.println("Average demand for best ants: " + (totalDemand/numberOfNodeCouples));
            return totalDemand/numberOfNodeCouples;
        }
    }
    public void shutdown() {
        graphDb.shutdown();
    }

    private static enum RelTypes implements RelationshipType
    {
        HAS_CONNECTION;

    }

    private static void registerShutdownHook( final GraphDatabaseService graphDb )
    {
        // Registers a shutdown hook for the Neo4j instance so that it
        // shuts down nicely when the VM exits (even if you "Ctrl-C" the
        // running application).
        Runtime.getRuntime().addShutdownHook( new Thread()
        {
            @Override
            public void run()
            {
                graphDb.shutdown();
            }
        } );
    }
}
